package li.klass.confluence.tableagg.plugin

import java.lang.String
import com.atlassian.confluence.pages.PageManager
import com.atlassian.confluence.`macro`.Macro
import com.atlassian.confluence.`macro`.Macro.{BodyType, OutputType}
import com.atlassian.confluence.content.render.xhtml.ConversionContext
import java.util.{Map => JMap}
import TableAggregationMacro._
import xml.{Node, NodeSeq, XML}


object TableAggregationMacro {
  private val MACRO_NAME = "tableagg"

  def aggregate(body: String, id: String, defaultValueIfText: Int, cellFunction: (String, NodeSeq) => Option[List[NodeSeq]], aggregationFunction: (Float, Float) => Float): Float = {
    val xml = toValidXML(body)
    val cells = cellFunction(id, xml)
    cells match {
      case None => -1
      case Some(foundCells) => {
        foundCells.map(cell => {
          try {
            cell.text.trim().toFloat
          } catch {
            case e: Exception => defaultValueIfText
          }
        }).reduceLeft(aggregationFunction(_, _))
      }
    }
  }

  def findColumnForId(id: String, xml: NodeSeq): Option[List[NodeSeq]] = {
    val tables = xml \\ "table"
    val tableResult = (for (table <- tables) yield {
      findVerticalRowsInTable(id, table)
    }).filter(_.isDefined).map(_.get)

    if (tableResult.size == 0) None else Some(tableResult.head)
  }

  def findVerticalRowsInTable(id: String, xml: NodeSeq): Option[List[NodeSeq]] = {
    val lines = xml \ "tbody" \ "tr"
    if (lines.size == 0) return None

    val columnCount = {
      val tdSize = (lines.head \ "td").size
      if (tdSize > 0) tdSize
      else (lines.head \ "th").size
    }

    val columns = for (i <- (0 to columnCount)) yield {
      val column: Seq[Node] = (for (line <- lines) yield {
        val cells = (line \ "td")
        if (! cells.isDefinedAt(i)) None
        else {
          Some(cells.apply(i))
        }
      }).filter(_.isDefined).map(_.get)
      column
    }

    val macroColumns = columns.filter(columnCells => {
      columnCells.exists(isSearchedMacroCell(id, _))
    })

    if (macroColumns.size == 0) None
    else Some(macroColumns.head.toList)
  }

  def findRowForId(id: String, xml: NodeSeq): Option[List[NodeSeq]] = {
    val tableRows: NodeSeq = xml \\ "table" \ "tbody" \ "tr"
    val validRows = tableRows.filter(isSearchedTableRow(id, _))
    
    if (validRows.size == 0) None else Some((validRows.head \ "td").toList)
  }

  def isSearchedTableRow(id: String, tableRow: NodeSeq): Boolean = {
    val cells = tableRow \ "td"
    cells.exists(isSearchedMacroCell(id, _))
  }

  def isSearchedMacroCell(id: String, cell: NodeSeq):  Boolean = {
    val macroCells = cell \\ "ac_macro"
    if (macroCells.size == 0) return false

    val macroCell = macroCells.head

    val macroName = (macroCell \ "@ac_name").head.text
    if (macroName != MACRO_NAME) return false;

    val parameters = macroCell \\ "ac_parameter"
    parameters.toList.exists(parameter => {
      val nameAttribute = parameter \ "@ac_name"
      if (nameAttribute.size == 0) false
      else if (nameAttribute.head.text != "id") false
      else if (parameter.text == id) true
      else false
    })
  }

  def toValidXML(content: String): NodeSeq = {
    val inputRegexp = "<input ([^>^/]*)>".r
    val inputValidatedBody = inputRegexp.replaceAllIn(content, "<input $1 />")
      .replaceAll("&nbsp;", "")
      .replaceAll("&[a-zA-Z]*;", "")
      .replaceAll("ac:macro", "ac_macro")
      .replaceAll("ac:parameter", "ac_parameter")
      .replaceAll("ac:name", "ac_name")
    XML.loadString("<div>" + inputValidatedBody + "</div>")
  }
}

class TableAggregationMacro(pageManager: PageManager) extends Macro {

  def getOutputType = OutputType.BLOCK

  def getBodyType = BodyType.NONE

  def execute(params: JMap[String, String], body: String, conversionContext: ConversionContext): String = {

    val body = conversionContext.getEntity.getBodyAsString
    val id = params.get("id")

    val aggregationParam = params.get("aggregation")
    val aggregationFunction: (Float, Float) => Float = aggregationParam match {
      case "sum" => _ + _
      case "multiply" => _ * _
      case "subtract" => _ - _
      case "divide" => _ / _
    }

    val defaultValueIfText = aggregationParam match {
      case "multiply" => 1
      case "divice" => 1
      case _ => 0
    }
    
    val orientation = params.get("orientation")
    val result = orientation match {
      case "horizontal" => aggregate(body, id, defaultValueIfText, findRowForId, aggregationFunction)
      case "vertical" => aggregate(body, id, defaultValueIfText, findColumnForId, aggregationFunction)
      case _ => -1
    }

    if (result == -1) {
      "??"
    } else {
       (scala.math.round(result * 100).toLong / 100f).toString
    }
  }
}
